from .misc import (
    cluster,
    merge_two_dicts,
    flatten,
    is_float,
    is_factor,
    check_and_install,
    is_multitask_estimator
)


__all__ = [
    "cluster",
    "merge_two_dicts",
    "flatten",
    "is_float",
    "is_factor",
    "check_and_install",
    "is_multitask_estimator"
]
